/*
 * Copyright 2012-2019 Intel Corporation
 *
 * This software and the related documents are Intel copyrighted materials, and
 * your use of them is governed by the express license under which they were
 * provided to you (License). Unless the License provides otherwise, you may
 * not use, modify, copy, publish, distribute, disclose, or transmit this
 * software or the related documents without Intel's prior written permission.
 *
 * This software and the related documents are provided as is, with no express
 * or implied warranties, other than those that are expressly stated in the
 * License.
 */

#ifndef _DS_DEFINES_H
#define _DS_DEFINES_H

#include <bitset>
#include <condition_variable>
#include <limits.h>
#include <map>
#include <memory>
#include <set>
#include <string>
#include <vector>

#include "provider.h"

struct _xmlNode;
using  xmlNode = _xmlNode;

namespace clck {

struct provider_config;

namespace datastore {

    namespace types {
      using exit_status_type = int32_t;
      using encoding_type    = int32_t;
      using node_count_type  = int32_t;
      using row_id_type      = int32_t;
      using version_type     = int32_t;
      using duration_type    = double;
      using time_type        = time_t;
      using host_name_type   = std::string;
      using optionid_type    = std::string;
      using provider_type    = std::string;
      using stdout_type      = std::string;
      using stderr_type      = std::string;
      using username_type    = std::string;
      using node_list_type   = std::vector<host_name_type>;
    }

    struct ParameterType {
    };

    struct Row : public ParameterType {
        types::row_id_type row_id        = -1;
        types::provider_type provider    = "";
        types::host_name_type hostname   = "";
        types::node_count_type num_nodes = -1;
        types::node_list_type node_names;
        types::exit_status_type exit_status = -1;
        types::time_type timestamp = 0;
        types::duration_type duration = 0;
        types::encoding_type encoding = 0;
        types::stdout_type stdout = "";
        types::stderr_type stderr = "";
        types::optionid_type OptionID = "";
        types::version_type version = 0;
        types::username_type username = "";
        types::time_type unique_timestamp = 0;
    };

    using Rows = std::vector<Row>;

    static std::map<std::string, int> colmap = {
        {"rowid", 0},
        {"Provider", 1},
        {"Hostname", 2},
        {"num_nodes", 3},
        {"node_names", 4},
        {"Exit_status", 5},
        {"Timestamp", 6},
        {"Duration", 7},
        {"Encoding", 8},
        {"STDOUT", 9},
        {"STDERR", 10},
        {"OptionID", 11},
        {"version", 12},
        {"Username", 13},
        {"Unique_timestamp", 14}
    };

    struct Row_baseline : public ParameterType {
        int rowid = -1;
        std::string baseline_id = "-1";
        int datastore_rowid = -1;
        std::string fwd_name = "-1";
        std::string provider_checksum = "-1";
        std::string command_checksum = "-1";
    };

    using Rows_baseline = std::vector<Row_baseline>;


    using Rows_dictionary_type = std::map<std::string, Rows>;

    using Snapshot_Rows_dictionary_type = std::map<std::string, Rows_baseline>;

    static std::map<std::string, int> colmap_baseline = {
        {"rowid", 0},
        {"baseline_id", 1},
        {"datastore_rowid", 2},
        {"fwd_name", 3},
        {"provider_checksum", 4},
        {"command_checksum", 5}
    };

    using datastore::types::row_id_type;

    using datastore::types::provider_type;

    /**
     * @brief Supported source types.
     */
    enum class Source_Type {
       //! type specification for a data source
       DATA = 0,
       //! type specification for log/streams
       STREAM,
       //! Type Count (DO NOT USE)
       SUPPORTED_COUNT
    };

    /**
     * @brief Supported datastore types.
     */
    enum class DatabaseType {
       //! Default sqlite3 type
       SQLITE = 0,
       //! Defaul ODBC type
       ODBC,
       //! User defined datastores
       OTHER
    };

    //! type for data source name
    using Datastore_Source_Name_t      = std::string;
    //!
    using Datastore_Instance_Name_t    = std::string;

    /**
     * @brief A datastore used for collection and analysis.
     */
    struct DataStoreConfig {
        //! Constructor
        DataStoreConfig() = default;
        DataStoreConfig(DatabaseType type, Datastore_Instance_Name_t i_name,
                        Datastore_Source_Name_t s_name, std::string d_path,
                        std::string sparams): type(type),instance_name(i_name),
                        datastoreExtensionName(s_name), datastoreExtensionPath(d_path),
                        source_parameters(sparams) {};
        //! Database type
        DatabaseType type = clck::datastore::DatabaseType::SQLITE;
        //! Database instance name
        Datastore_Instance_Name_t instance_name = "";
        //! Database source (library) name
        Datastore_Source_Name_t datastoreExtensionName = "";
        //! datasource extension path
        std::string datastoreExtensionPath = "";
        //! Database source parameters
        std::string source_parameters = "";
        //! Database source type classification
        std::bitset<(size_t)Source_Type::SUPPORTED_COUNT> source_types = 0;
    };

    using DataStoreCfgList_type = std::vector<DataStoreConfig>;

    struct SQLParameters : ParameterType {
       //! where clause (specifically for default SQLite and ODBC datastore(s))
       std::string where_clause = "";
       //! Mutually exclusive with everything except colMap
       std::string query = "";
       //! Coloumn maping (specifically for default SQLite and ODBC datastore(s))
       std::map<std::string, int> colMap = clck::datastore::colmap;
    };

    struct ReadParameters : public ParameterType {
       //!  list of providers to be loaded
       std::vector<std::string> providers = {};
       //!  list of hosts to be loaded
       std::vector<std::string> hostnames = {};
       //!  list of baselines to be loaded
       std::vector<std::string> selected_baselines = {};
       //! Additional parameters (if required)
       ParameterType *secondary_parameters = nullptr;
    };

    struct WriteData : public ParameterType {
       //! stdout to be written to the datastore
       std::string stdout;
       //! stderr to be written to the datastore
       std::string stderr;
       //! Metadata
       clck::provider::header header;
       //! Node names for which the data was generated
       std::string node_names;
       //! Additional parameters (if required)
       ParameterType *secondary_parameters = nullptr;
    };

    // TODO Vish -  Fix the data structure
    struct ProviderInfo {
      row_id_type rowid = 0;
      std::string fwd_name;
      std::shared_ptr<clck::provider_config> pconfig;
    };

    using ProviderInfoMapping = std::map<provider_type, ProviderInfo>;

    // Use to compare the set key of struct provider_info
    // based on provider row id
    struct setCompareProviderRowId {
        bool operator() (const ProviderInfo& lhs,
            const ProviderInfo& rhs) const {
            return lhs.rowid < rhs.rowid;
        }
    };

    using SnapshotTableFieldsMappingType = std::map<std::string, std::set<ProviderInfo, setCompareProviderRowId>>;

    struct SnapshotUpdateData : public ParameterType {
       //! snapshot dictionary (baseline id is the key)
       SnapshotTableFieldsMappingType* snapshot_dict = nullptr;
    };

    struct SnapshotInfo {
      std::string baseline_id;
      // Timestamp when data was collected
      types::time_type collect_timestamp = 0;
    };

    // Use to compare the baseline id of struct snapshot_info
    struct setCompareBaselineId {
        bool operator() (const SnapshotInfo& lhs,
            const SnapshotInfo& rhs) const {
            return lhs.baseline_id < rhs.baseline_id;
        }
    };

    using SnapshotInfoSet = std::set<SnapshotInfo, setCompareBaselineId>;

    enum class SUPPORTED_OPS {
      //! operational codes upto 512 are internal and should not be used
      //! should be able to return the hosts present in the datastore
      GET_HOSTS = 512,
      //! should be able to return the rows corresponding to latest timestamp
      LATEST_TIMESTAMP_READ,
      //! operation to get the corresponding data to a provider
      READ,
      //! standard write operation using the WriteData struct
      WRITE,
      //! baseline query operation (to query available baselines)
      //! Enabled only for default SQLite type
      GET_BASELINES,
      //! queries the snapshots for provided baselines
      //! Enabled only for default SQLite type
      SNAPSHOT_READ,
      //! commits the data to create a new snapshot
      //! Enabled only for default SQLite type
      SNAPSHOT_UPDATE,
      // Marks the latest collected rows
      // Enabled only for default SQLite type
      SNAPSHOT_MARK_LATEST,
      // removes the provided baseline
      // Enabled only for default SQLite type
      SNAPSHOT_REMOVE
    };
};
};

#endif
