/* 
 * Copyright 2012-2019 Intel Corporation
 *
 * This software and the related documents are Intel copyrighted materials, and
 * your use of them is governed by the express license under which they were
 * provided to you (License). Unless the License provides otherwise, you may
 * not use, modify, copy, publish, distribute, disclose, or transmit this
 * software or the related documents without Intel's prior written permission.
 *
 * This software and the related documents are provided as is, with no express
 * or implied warranties, other than those that are expressly stated in the
 * License.
 */
/**
 * @file
 * @brief Database Interface class.
 */

#ifndef _DATASTORE_EXTENSION_H
#define _DATASTORE_EXTENSION_H

#include <config/datastore_defines.h>

namespace clck {

namespace datastore {

//Database interface base class.
class DataStoreExtension {
 public:
    //! Error descriptor
    using Datastore_error_descriptor_t = int32_t;

    //! Error descriptor
    using Datastore_opcode_t           = uint16_t;

    //! return type for data store extensions
    using DataStoreExtension_t         = const std::weak_ptr<DataStoreExtension>;

    //! Error code for returning unsupported operation
    static const Datastore_error_descriptor_t UNSUPPORTED_OPERATION;

    //! Error code for returning unsupported operation
    static const Datastore_error_descriptor_t OPERATION_SUCCESSFULL;

    //! Error code for returning unsupported operation
    static const Datastore_error_descriptor_t OPERATION_UNSUCCESSFULL;

    //! returns the type of data store extension
    DatabaseType getType() {
      return type;
    }

    //! returns the name of data store extension
    std::string get_name() {
      return name;
    }

    //! sets the name of data store extension
    void set_name (Datastore_Source_Name_t name) {
      this->name = name;
    }

    uint64_t get_version() {
      return version;
    }

    //! defaulted virtual destructor
    virtual ~DataStoreExtension() = default;

   friend class DataStoreController;

    //! disabling all the by value copy/move
    DataStoreExtension& operator=(const DataStoreExtension&) = delete;
    DataStoreExtension& operator=(DataStoreExtension&&) = delete;
    DataStoreExtension(const DataStoreExtension&) = delete;
    DataStoreExtension(DataStoreExtension&&) = delete;
protected:
    using super = DataStoreExtension;
    //! current version of extension
    uint64_t version = 20190;
    //! Function to return if connection is open
    virtual bool is_open() = 0;
    //! generic function to define operations on data store extensions
    virtual Datastore_error_descriptor_t control(Datastore_opcode_t opcode, const void* parameters, void* data, void* status) = 0;
    //! Name
    Datastore_Source_Name_t name = "default-datatstore-extension";
    //! Disable the default Ctor
    DataStoreExtension() = delete;
    //! Allowed Ctor
    explicit DataStoreExtension(DatabaseType type): type(type) {
    }
    //! Function to open the datastore
    virtual Datastore_error_descriptor_t open(const std::string &parameters) = 0;
    //! Function to close the data store
    virtual Datastore_error_descriptor_t close() = 0;

private:
    //! DatabaseType
    DatabaseType type;
};

typedef DataStoreExtension *CreateDSExtension();

typedef void DestoryDSExtension(DataStoreExtension *obj);


};
};

#endif
