/*
 * Copyright 2012-2019 Intel Corporation
 *
 * This software and the related documents are Intel copyrighted materials, and
 * your use of them is governed by the express license under which they were
 * provided to you (License). Unless the License provides otherwise, you may
 * not use, modify, copy, publish, distribute, disclose, or transmit this
 * software or the related documents without Intel's prior written permission.
 *
 * This software and the related documents are provided as is, with no express
 * or implied warranties, other than those that are expressly stated in the
 * License.
 */

/**
 * @file
 * @brief This file contains the public interface to the daemon datastore class.
 */

#ifndef _DATABASE_H
#define _DATABASE_H

#include <map>
#include <memory>
#include <string>
#include <vector>
#include <config/datastore_defines.h>
#include <datastore/data_store_extensions.h>

namespace clck {
	namespace datastore {

        using CallBack = void (*) (DataStoreExtension::Datastore_error_descriptor_t,
                                   uint16_t,
                                   DataStoreExtension::DataStoreExtension_t,
                                   const void *data,
                                   const void *status
                                  );

        //!
        //! @brief  This wrapper method loops over opened data stores and executes the query on the data sources
        //!         Works only for SQL type data sources
        //! @param  query Query to be executed on SQL data sources
        //! @param  columns Column mapping for the parameters to be queried
        //! @param  mark_as_baseline set to be true if queried data is to be used for baseline creation
        //! @return Rows Queried rows
        //! @return bool returns false only if querying is aborted
        //!
        bool select_data(const std::string query, Rows& rows,
                         const std::map<std::string,
                         int>& columns=colmap, bool mark_as_baseline = false);


        //!
        //! @brief  This wrapper method loops over opened data stores and gets the latest timestamp data
        //!         from the first data source it can find the data.
        //! @param  providers non empty list of providers to be loaded
        //! @param  hostname particular hostname to be found (optional)
        //! @param  mark_as_baseline set to be true if queried data is to be used for baseline creation
        //! @return Rows Queried rows
        //! @return bool returns false only if querying is aborted
        //!

        bool get_latest_rows_provider_data(const std::vector<std::string> &providers,
                                           Rows& rows, std::string hostname="",
                                           bool mark_as_baseline = false);

        //!
        //! @brief  This wrapper method loops over opened data stores and gets the latest timestamp data
        //!         from the first data source it can find the data.
        //! @param  providers non empty list of providers to be loaded
        //! @param  hostnames list of hostnames to be found (optional)
        //! @param  mark_as_baseline set to be true if queried data is to be used for baseline creation
        //! @return Rows Queried rows
        //! @return bool returns false only if querying is aborted
        //!

        bool get_latest_rows_provider_hosts_data(const std::vector<std::string> &providers,
                                           Rows& rows, const std::vector<std::string>& hostnames,
                                           bool mark_as_baseline = false);

        //!
        //!
        //! @brief  This wrapper method loops over opened data stores and gets hosts available from the first
        //!         data source it can find the data.
        //! @param  columns Column mapping for the parameters to be queried
        //! @param  mark_as_baseline set to be true if queried data is to be used for baseline creation
        //! @return Rows Queried rows
        //! @return bool returns false only if querying is aborted
        //!
        bool get_hosts_data(Rows& rows, const std::map<std::string, int>& columns=colmap
                        ,bool mark_as_baseline=false);

        //!
        //! @brief  This wrapper method loops over opened data stores and gets data points
        //!         for the requested snapshot from the first data source it can
        //!         find the data.
        //! @param  provider provider name for which basline data needs to be extracted
        //! @param  mark_as_baseline set to be true if queried data is to be used for baseline creation
        //! @return b_rows_vec Queried rows
        //! @return s_rows_vec Queried rows
        //! @return bool returns false only if querying is aborted
        //!
        bool get_snapshot_provider_data(std::string provider,
                                    Rows_dictionary_type &b_rows_vec,
                                    Snapshot_Rows_dictionary_type &s_rows_vec,
                                    const std::vector<std::string> hostnames={},
                                    bool mark_as_baseline = false);

        //!
        //! @brief  This wrapper method loops over opened data stores and gets provider(s) data
        //!         available from the first data source it can find the data.
        //! @param  providers non empty list of providers to be loaded
        //! @param  where_clause where clause (optional, works only for default types SQL data sources)
        //! @param  mark_as_baseline set to be true if queried data is to be used for baseline creation
        //! @return Rows Queried rows
        //! @return bool returns false only if querying is aborted
        //!
        bool select_provider_data(std::vector<std::string> providers, Rows& rows,
                                  std::string where_clause="",
                                  bool mark_as_baseline = false);
        //!
        //! @brief  This wrapper method takes a list of data sources as operand (of the data sources loaded)
        //! @param  data_sources non empty list of data source names (operands)
        //! @param  opcode Operation code of operation to be executed
        //! @param  controlParams control parameters for the operation
        //! @param  cb Callback which is called if ret status is anything but success.
        //! @return data queried data
        //! @return status Status handle (if returned by the individual data store extension)
        //! @return bool returns false only if querying is aborted
        //!
        bool execute_ioctl(std::vector<std::string> data_sources,
                                  uint16_t opcode,
                                  const void* controlParams,
                                  void* data,
                                  void* status, CallBack cb = nullptr);


    };  // namespace datastore
};  // namespace clck

#endif // _DATABASE_H
