 ; Copyright 2006-2019 Intel Corporation
 ;
 ; This software and the related documents are Intel copyrighted materials, and
 ; your use of them is governed by the express license under which they were
 ; provided to you (License). Unless the License provides otherwise, you may
 ; not use, modify, copy, publish, distribute, disclose, or transmit this
 ; software or the related documents without Intel's prior written permission.
 ;
 ; This software and the related documents are provided as is, with no express
 ; or implied warranties, other than those that are expressly stated in the
 ; License.

(defclass CPU_TRAITS
  "This class represents the CPU types on a system."
  (is-a USER)
  (role abstract)
  (slot cpu-type (type SYMBOL) (allowed-values generic xeonphi xeonphi_m xeon_scalable)
                                (default generic) (visibility public)))

(defclass CPU_KERNEL_TRAITS
  "This class represents the kernel configurable properties w.r.t to CPU
  on a system."
  (is-a USER)
  (role abstract)
  ; user specified cpu isolation list
  (slot cpu-list-isol (type STRING) (default "") (visibility public))
  ; user specified cpu no callbacks list
  (slot cpu-list-ncb (type STRING) (default "") (visibility public))
  ; dmesg error stream
  (slot dmesg-stderr (type STRING) (default "") (visibility public))
  ; kernel error msg
  (slot kernel-nohz-errormsg (type STRING) (default "") (visibility public))
  ; user specified tickless cpu list
  (slot tickless-list-usr (type STRING) (default "") (visibility public))
  ; kernel accepted tickless cpu list
  (slot tickless-list-kernel (type STRING) (default "") (visibility public)))

(defclass CPU_CORE_TRAITS
  "This class represents the CPU cores on a system."
  (is-a USER)
  (role abstract)
  ;Available logical cores for a given CPU
  (slot available-logical-cores (type INTEGER) (default 0) (visibility public))
  ;Available physical cores for a given CPU socket
  (slot available-physical-cores (type INTEGER) (default 0) (visibility public))
  ;max physical cores for a given CPU (currently unused)
  (slot max-phyiscal-cores (type INTEGER) (default 0) (visibility public))
  ;Max logical cores for a given CPU
  (slot max-logical-cores (type INTEGER) (default 0) (visibility public))
  ;max threads per core for a given CPU (currently unused)
  (slot max-threads-per-core (type INTEGER) (default 0) (visibility public))
  ;Offline logical cores for a given CPU
  (slot offline-logical-cores (type STRING) (visibility public) (default ""))
  ;number of sockets on the system
  (slot sockets (type INTEGER) (default 0) (visibility public))
  (slot expected-sockets (type INTEGER) (default 0) (visibility public))
  ;valid threads per core
  (multislot valid-threads-per-core (storage shared) (type INTEGER)
                                    (access read-only)
                                    (default (create$ 1 2 4)))
  ;available threads per core for a given CPU
  (slot threads-per-core (type INTEGER) (default 0) (visibility public))
)

(defclass CPU
  "This class represents the CPU on a system."
  (is-a BASE_NODE BASE_TIMESTAMP DATABASE MULTISET CPU_TRAITS CPU_CORE_TRAITS
        CPU_KERNEL_TRAITS)
  (role concrete)
  (pattern-match reactive)

  (slot cpuid-str (type STRING) (default ""))
  ; cpuid error stream
  (slot cpuid-stderr (type STRING) (default ""))
  ; cpuinfo error stream
  (slot cpuinfo-stderr (type STRING) (default ""))
  (slot family (type INTEGER) (default 0))
  (multislot flags (type SYMBOL))
  ; intel_pstate_status error stream
  (slot intel-pstate-stderr (type STRING) (default ""))
  ; dmesg error stream
  (slot lscpu-stderr (type STRING) (default ""))
  (slot model (type INTEGER) (default 0))
  (slot model-name (type STRING) (default ""))
  (slot expected-model-name (type STRING) (default ""))
  (multislot req-flags (type SYMBOL))
  (slot stepping (type INTEGER) (default 0))
  (slot siblings (type INTEGER) (default 0))
  (slot turbo-status (type SYMBOL) (allowed-symbols enabled disabled nil)
                                   (default nil))
  (slot vendor-id (type STRING) (default "")))

(defmessage-handler CPU init after ()
  (send ?self add (send ?self uniformity-key model-name) ?self:model-name)
  (send ?self add (send ?self uniformity-key tickless-list-usr)
                   ?self:tickless-list-usr)
  (send ?self add (send ?self uniformity-key turbo-status) ?self:turbo-status)
  (send ?self add (send ?self uniformity-key-model-based available-logical-cores)
                   ?self:available-logical-cores)
  (send ?self add (send ?self uniformity-key-model-based threads-per-core)
                   ?self:threads-per-core))

(defmessage-handler CPU uniformity-key (?v)
  "Return a key that is unique for each combination of subcluster,
   role, and architecture."

  (bind ?role nil)
  (if (member$ compute ?self:role) then (bind ?role compute))
  (if (member$ enhanced ?self:role) then (bind ?role enhanced))
  (bind ?key (sym-cat (class ?self) + ?self:subcluster + ?self:architecture +
                      ?role + ?v))
  (return ?key))

(defmessage-handler CPU uniformity-key-model-based (?v)
  "Return a key that is unique for each combination of subcluster,
   role, and architecture."
  (bind ?role nil)
  (if (member$ compute ?self:role) then (bind ?role compute))
  (if (member$ enhanced ?self:role) then (bind ?role enhanced))
  (bind ?key (sym-cat (class ?self) + ?self:subcluster + ?self:architecture +
                      ?role + ?self:cpu-type + ?self:model-name + ?v))
  (return ?key))


(defclass XP_MODES
  "Mode class for the Intel(R) Xeon Phi(TM) processor"
  (is-a  BASE_NODE BASE_TIMESTAMP DATABASE CPU_TRAITS)
  (role concrete)
  (pattern-match reactive)

  (slot cluster-mode (type SYMBOL) (allowed-values nil All2All Quadrant Hemishpere
                                                   SNC2 SNC4 non-numa))
  (slot memory-mode (type SYMBOL) (allowed-values nil Cache Flat Hybrid25
                                                  Hybrid50 numa))
  (slot last-modified (type INTEGER) (default 0))
  (slot last-reboot (type INTEGER) (default 0))
  ; source provider used for filling up the fields
  (multislot data-source (type SYMBOL) (allowed-values hwloc_dump_hwdata numactl))
  (slot stderr (type STRING) (default ""))
  (slot numactl-stderr (type STRING) (default "")))

(defmessage-handler XP_MODES init after ()
  (multiset-insert (send ?self uniformity-key cluster-mode) ?self:cluster-mode)
  (multiset-insert (send ?self uniformity-key memory-mode) ?self:memory-mode))

(defmessage-handler XP_MODES uniformity-key (?mode)
  (bind ?role other)
  (if (member$ compute ?self:role) then (bind ?role compute))
  (if (member$ enhanced ?self:role) then (bind ?role enhanced))
  (bind ?key (sym-cat (class ?self) + ?self:subcluster + ?self:architecture +
                      ?role + ?self:cpu-type + ?mode))
  (return ?key))
