 ; Copyright 2006-2019 Intel Corporation
 ;
 ; This software and the related documents are Intel copyrighted materials, and
 ; your use of them is governed by the express license under which they were
 ; provided to you (License). Unless the License provides otherwise, you may
 ; not use, modify, copy, publish, distribute, disclose, or transmit this
 ; software or the related documents without Intel's prior written permission.
 ;
 ; This software and the related documents are provided as is, with no express
 ; or implied warranties, other than those that are expressly stated in the
 ; License.

(defclass MEMORY_MODULE_TRAITS
  "This class represents memory modules on a local system."
  (is-a BASE_TIMESTAMP DATABASE STATISTICS)
  (role abstract)
  (slot source (type SYMBOL) (default DMIDECODE))
  ; DIMM indices (current working solution)
  (multislot dimm-indices (type INTEGER) (default 0) (range 0 ?VARIABLE) (visibility public))
  (multislot index (type INTEGER) (default 0) (range 0 ?VARIABLE) (visibility public))
  ; DIMM Sizes
  (multislot dimm-sizes (type INTEGER) (default 0) (range 0 ?VARIABLE) (visibility public))
  ; DIMM Speeds
  (multislot dimm-speeds (type INTEGER) (default 0) (range 0 ?VARIABLE) (visibility public))
  ; DIMM Form Factors
  (multislot dimm-form-factors (type STRING) (visibility public))
  ; DIMM Types
  (multislot dimm-types (type STRING) (visibility public))
  ; DIMM locators
  (multislot dimm-locators (type STRING) (visibility public))
  ; DIMM Count
  (slot dimm-count (type INTEGER) (default 0) (visibility public))
  ; DIMM Slot Count
  (slot max-dimm-count (type INTEGER) (default 0) (visibility public))
  ; Minimal dimms per socket
  (slot min-dimms-per-socket (type INTEGER) (default 0) (visibility public))
  (multislot dimms-per-socket (type INTEGER) (default 0) (visibility public))
  (slot dmidecode-path-err (type STRING) (visibility public))
  (slot dmidecode-err (type STRING) (visibility public))
)

(defclass MEMORY
  "This class represents memory on a local system."
  (is-a BASE_NODE MEMORY_MODULE_TRAITS)
  (role concrete)
  (pattern-match reactive)

  (slot physical-free (type INTEGER) (default 0) (range 0 ?VARIABLE))
  (slot physical-per-core (type INTEGER) (default 0) (range 0 ?VARIABLE))
  (slot physical-total (type INTEGER) (default 0) (range 0 ?VARIABLE))
  (slot expected-physical-total (type INTEGER) (default 0) (range 0 ?VARIABLE))
  (slot expected-physical-per-core (type INTEGER) (default 0) (range 0 ?VARIABLE))
  (slot swap-free (type INTEGER) (default 0) (range 0 ?VARIABLE))
  (slot swap-total (type INTEGER) (default 0) (range 0 ?VARIABLE)))

(defmessage-handler MEMORY init after ()
  (bind ?update-physical-memory FALSE)
  (if (=  ?self:physical-total 0) then
    (bind ?update-physical-memory TRUE))
  (progn$ (?dimm-size ?self:dimm-sizes)
    ;insert for uniformity only if memory installed on the DIMM slot
    (if (> ?dimm-size 0) then
      (multiset-insert (send ?self dimm-uniformity-key
                       (nth$ ?dimm-size-index ?self:dimm-form-factors) SIZE)
                       ?dimm-size)
      (multiset-insert (send ?self dimm-uniformity-key
                       (nth$ ?dimm-size-index ?self:dimm-form-factors) TYPE)
                       (nth$ ?dimm-size-index ?self:dimm-types))
      (multiset-insert (send ?self dimm-uniformity-key
                       (nth$ ?dimm-size-index ?self:dimm-form-factors) SPEED)
                       (nth$ ?dimm-size-index ?self:dimm-speeds))
        (if (eq ?update-physical-memory TRUE) then
          (bind ?self:physical-total (+ ?self:physical-total ?dimm-size)))))
  (progn$ (?dimms-count ?self:dimms-per-socket)
      (multiset-insert (send ?self node-based-uniformity-key DIMMPS) ?dimms-count)
  )
    ;insert for uniformity only if memory installed on the DIMM slot
  (send ?self add (send ?self statistics-key physical-total)
        ?self:physical-total))

(defmessage-handler MEMORY statistics-key (?v)
  (bind ?role nil)
  (bind ?subcluster nil)
  (bind ?arch nil)
  (bind ?ins (find-instance ((?n NODE)) (eq ?n:node_id ?self:node_id)))
  (if (= (length ?ins) 1) then
    (bind ?i (nth$ 1 ?ins))
    (bind ?arch (send ?i get-architecture))
    (bind ?subcluster (send ?i get-subcluster))
    (if (member$ compute (send ?i get-role)) then
      (bind ?role compute)
      else (if (member$ enhanced (send ?i get-role)) then
             (bind ?role enhanced))))

  (bind ?key (sym-cat (class ?self) + ?subcluster + ?role + ?arch + ?v))
  (return ?key))

; generate a form factor and node id based uniformity key
(defmessage-handler MEMORY dimm-uniformity-key (?formfactor ?v)
  "Return a key that is unique for each combination of subcluster,
   role, and architecture."
  (bind ?role nil)
  (if (member$ compute ?self:role) then (bind ?role compute))
  (if (member$ enhanced ?self:role) then (bind ?role enhanced))
  (bind ?key (sym-cat (class ?self) + ?self:subcluster + ?self:architecture +
                      ?role + ?self:node_id + ?formfactor + ?v))
  (return ?key))

; generate a form factor and node id based uniformity key
(defmessage-handler MEMORY node-based-uniformity-key (?v)
  "Return a key that is unique for each combination of subcluster,
   role, and architecture."
  (bind ?role nil)
  (if (member$ compute ?self:role) then (bind ?role compute))
  (if (member$ enhanced ?self:role) then (bind ?role enhanced))
  (bind ?key (sym-cat (class ?self) + ?self:subcluster + ?self:architecture +
                      ?role + ?self:node_id + ?v))
  (return ?key))

(defmessage-handler MEMORY get-dimm-size-by-idx(?idx)
  (return (nth$ ?idx ?self:dimm-sizes)))

(defmessage-handler MEMORY get-dimm-type-by-idx(?idx)
  (return (nth$ ?idx ?self:dimm-types)))

(defmessage-handler MEMORY get-dimm-speed-by-idx(?idx)
  (return (nth$ ?idx ?self:dimm-speeds)))

(defmessage-handler MEMORY get-form-factor-by-idx(?idx)
  (return (nth$ ?idx ?self:dimm-form-factors)))

(defmessage-handler MEMORY get-locator-by-idx(?idx)
  (return (nth$ ?idx ?self:dimm-locators)))
