 ; Copyright 2006-2019 Intel Corporation
 ;
 ; This software and the related documents are Intel copyrighted materials, and
 ; your use of them is governed by the express license under which they were
 ; provided to you (License). Unless the License provides otherwise, you may
 ; not use, modify, copy, publish, distribute, disclose, or transmit this
 ; software or the related documents without Intel's prior written permission.
 ;
 ; This software and the related documents are provided as is, with no express
 ; or implied warranties, other than those that are expressly stated in the
 ; License.

(defclass RPM
  (is-a BASE_NODE BASE_TIMESTAMP DATABASE MULTISET)
  (role concrete)
  (pattern-match reactive)

  ; The rpms and state slots are paired.  The nth value of the state
  ; slot corresponds to the nth value of the rpms slot.  The state
  ; slot may contain the value 'present' denoting that the
  ; corresponding rpm is present on all nodes, 'missing' denoting that
  ; the corresponding rpm is missing from the corresponding node but
  ; is present on at least one other node, and 'extra' denoting that
  ; the corresponding rpm is present on the corresponding node but is
  ; missing on at least one other node.
  ; In order to verify that a particular rpm is installed, check that
  ; the rpm has either the 'present' or the 'extra' state, or
  ; equivalently, does not have the 'missing' state.
  (multislot rpms (type SYMBOL))
  (multislot state (type SYMBOL) (allowed-values extra missing present))
  (multislot names (type SYMBOL))
  (multislot versions (type STRING))
  (slot infiniband-diags-version (type SYMBOL))
  (slot libfabric-version (type SYMBOL))
  (slot rdma-core-version (type SYMBOL))
  (slot embree2 (type STRING) (default "none"))
  (slot embree3 (type STRING) (default "none"))

  ; SSF RPMs
  (multislot req-rpms (storage shared) (type SYMBOL) (access read-only)
      (default (create$ gcc intel-icc-runtime-64bit intel-ifort-runtime-64bit
                       intel-mkl-runtime-64bit intel-tbb-runtime-64bit
                       intel-mpi-runtime-64bit)))

  ; SSF RPM minimum versions
  (multislot req-versions (storage shared) (type STRING) (access read-only)
      (default (create$ "4.8" "18.0" "18.0" "2018.0" "2018" "2018"))))

(defmessage-handler RPM init after ()
  ; Handle head nodes that are also compute nodes differently since
  ; they are allowed to have extra rpms that are not present on
  ; dedicated compute nodes.
  (if (and (member$ head ?self:role)
           (or (member$ compute ?self:role) (member$ enhanced ?self:role)))
    then
    ; insert each rpm key individually, skipping the extra ones
    (progn$ (?rpm ?self:rpms)
            (bind ?state (send ?self find-state ?rpm))
            (if (neq ?state extra) then (multiset-insert (send ?self uniformity-key ?rpm)
                                                         ?state)))
    else
    ; for all other nodes, do a bulk insert
    (multiset-insert-multifield (send ?self uniformity-key "") ?self:rpms
                                ?self:state))

    ; Set versions
    (send ?self set-versions)

    ; infiniband-diags
    (if (neq (member$ infiniband-diags ?self:names) FALSE)
        then (send ?self add
             (send ?self uniformity-key-single infiniband-diags-version)
             ?self:infiniband-diags-version))

    ; libfabric
    (if (neq (member$ libfabric ?self:names) FALSE)
        then (send ?self add
             (send ?self uniformity-key-single libfabric-version)
             ?self:libfabric-version))

    ; rdma-core
    (if (neq (member$ rdma-core ?self:names) FALSE)
        then (send ?self add
             (send ?self uniformity-key-single rdma-core-version)
             ?self:rdma-core-version)))

(defmessage-handler RPM uniformity-key (?rpm)
  (bind ?role other)
  (if (member$ compute ?self:role) then (bind ?role compute))
  (if (member$ enhanced ?self:role) then (bind ?role enhanced))
  (bind ?key (sym-cat (class ?self) + ?self:subcluster + ?self:architecture +
                      ?role + ?rpm))
  (return ?key))

(defmessage-handler RPM uniformity-key-single (?v)
  (bind ?key (sym-cat
    (class ?self) + ?self:subcluster + ?self:architecture + ?v))
  (return ?key))

(defmessage-handler RPM find-state (?rpm)
  (bind ?i (member$ ?rpm ?self:rpms))
  (return (nth$ ?i ?self:state)))

(defmessage-handler RPM find-version (?rpm)
  (bind ?i (member$ ?rpm ?self:names))
  (return (nth$ ?i ?self:versions)))

(defmessage-handler RPM find-req-version (?rpm)
  (bind ?i (member$ ?rpm ?self:req-rpms))
  (return (nth$ ?i ?self:req-versions)))

(defmessage-handler RPM set-versions ()

  ; infiniband-diags
  (if (neq (member$ infiniband-diags ?self:names) FALSE)
      then (bind ?self:infiniband-diags-version
           (send ?self find-version infiniband-diags)))

  ; libfabric
  (if (neq (member$ libfabric ?self:names) FALSE)
      then (bind ?self:libfabric-version
           (send ?self find-version libfabric)))

  ; rdma-core
  (if (neq (member$ rdma-core ?self:names) FALSE)
      then (bind ?self:rdma-core-version
           (send ?self find-version rdma-core))))
