 ; Copyright 2006-2019 Intel Corporation
 ;
 ; This software and the related documents are Intel copyrighted materials, and
 ; your use of them is governed by the express license under which they were
 ; provided to you (License). Unless the License provides otherwise, you may
 ; not use, modify, copy, publish, distribute, disclose, or transmit this
 ; software or the related documents without Intel's prior written permission.
 ;
 ; This software and the related documents are provided as is, with no express
 ; or implied warranties, other than those that are expressly stated in the
 ; License.

(defclass BASE_TIMESTAMP
  "Base class for data point timestamps.  This class should be
   inherited by all classes that need to have a timestamp slot."
  (is-a USER)
  (role abstract)

  ; seconds since epoch
  (slot timestamp (type NUMBER) (default 0) (visibility public)))

(defclass TIMESTAMP
  "This class should be need to contain logic that requires knowledge
  of which of multiple data points is the latest.

  Note: ideally the rule that sets the latest-timestamp slot would be
  defined in this file.  However, as described in section 5.4.1.7 of
  the CLIPS Basic Programming Guide, this rule needs to be defined
  after all the files in classes/ have been loaded."
  (is-a BASE_TIMESTAMP)
  (role abstract)

  (multislot _timestamp_keys (type SYMBOL)
             (pattern-match non-reactive) ; critical for performance
             (storage shared)
             (visibility private))
  (multislot _maximum_timestamp (type NUMBER)
             (pattern-match non-reactive) ; critical for performance
             (storage shared)
             (visibility private))
  (slot latest-timestamp (type SYMBOL) (allowed-values TRUE FALSE)
        (default FALSE)))

(defmessage-handler TIMESTAMP init after ()
  "Keep track of the maximum timestamp."
  (bind ?key (sym-cat (class ?self) - (send ?self get-node_id)))
  (bind ?pos (member$ ?key ?self:_timestamp_keys))
  (if ?pos then
    ; key/value pair already exists
    (bind ?current-max (nth$ ?pos ?self:_maximum_timestamp))
    (if (> ?self:timestamp ?current-max) then
      (slot-replace$ ?self _maximum_timestamp ?pos ?pos ?self:timestamp))
    else
    ; key/value pair needs to be created
    (slot-insert$ ?self _timestamp_keys 1 ?key)
    (slot-insert$ ?self _maximum_timestamp 1 ?self:timestamp)))

(defmessage-handler TIMESTAMP get-maximum-timestamp ()
  "Return the maximum timestamp."
  (bind ?key (sym-cat (class ?self) - (send ?self get-node_id)))
  (bind ?pos (member$ ?key ?self:_timestamp_keys))
  (if ?pos
      then (return (nth$ ?pos ?self:_maximum_timestamp))
      else (return FALSE)))

(defrule set-latest-timestamp
  "Set latest-timestamp to TRUE for those instances that correspond to
   the maximum timestamp."
  (declare (salience ?*first-salience*))
  =>
  (do-for-all-instances ((?i TIMESTAMP))
                        (>= ?i:timestamp (send ?i get-maximum-timestamp))
                        (send ?i put-latest-timestamp TRUE)))
