; Copyright 2006-2019 Intel Corporation
;
; This software and the related documents are Intel copyrighted materials, and
; your use of them is governed by the express license under which they were
; provided to you (License). Unless the License provides otherwise, you may
; not use, modify, copy, publish, distribute, disclose, or transmit this
; software or the related documents without Intel's prior written permission.
;
; This software and the related documents are provided as is, with no express
; or implied warranties, other than those that are expressly stated in the
; License.

(defrule substandard-dgemm-due-to-high-cpu-process
  "Diagnose substandard DGEMM performance issues due to a conflicting
   process that is consuming a high amount of cpu.  If the
   dgemm-performance sign is substandard and the high-cpu-process sign
   is true and the associated data points are close together in
   time (within 10 minutes), then conclude the substandard performance
   is due to the high cpu process."
  ?sign1 <- (object (is-a SIGN) (node_id ?node)
                    (id "dgemm-data-is-substandard") (source ?dgemm)
                    (suppressed FALSE) (value substandard))
  ?sign2 <- (object (is-a SIGN) (node_id ?node) (id "process-is-high-cpu")
                    (source ?process) (suppressed FALSE) (value TRUE))
  (test (< (abs (- (send ?dgemm get-timestamp)
                   (send ?process get-timestamp))) 600))
  =>
  ; the further apart the two data points in time, the lower the confidence
  (bind ?confidence (combine-confidence
                     (send ?sign1 get-confidence)
                     (send ?sign2 get-confidence)
                     (* (- 1 (/ (abs (- (send ?dgemm get-timestamp)
                                        (send ?process get-timestamp)))
                                600)) 100)))
  (bind ?severity (combine-severity (send ?sign1 get-severity)
                                    (send ?sign2 get-severity)))
  (make-instance of DIAGNOSIS (node_id ?node) (confidence ?confidence)
                 (severity ?severity)
                 (id "substandard-dgemm-due-to-high-cpu-process")
                 (args (create$ (/ (send ?dgemm get-flops) 1e9)
                                (send ?process get-pid)))
                 (remedy "process-is-high-cpu-remedy")
                 (remedy-args (create$ (send ?process get-pid))))
  (modify-instance ?sign1 (state diagnosed))
  (modify-instance ?sign2 (state diagnosed)))

(defrule substandard-dgemm-due-to-high-memory-process
  "Diagnose substandard DGEMM performance issues due to a conflicting
   process that is consuming a large amount of memory.  If the
   dgemm-performance sign is substandard and the high-memory-process
   sign is true and the associated data points are close together in
   time (within 10 minutes), then conclude the substandard performance
   is due to the high memory process."
  ?sign1 <- (object (is-a SIGN) (node_id ?node)
                    (id "dgemm-data-is-substandard") (source ?dgemm)
                    (suppressed FALSE) (value substandard))
  ?sign2 <- (object (is-a SIGN) (node_id ?node) (args $?args)
                    (id "process-is-high-memory") (source ?process)
                    (suppressed FALSE) (value TRUE))
  (test (< (abs (- (send ?dgemm get-timestamp)
                   (send ?process get-timestamp))) 600))
  =>
  ; the further apart the two data points in time, the lower the confidence
  (bind ?confidence (combine-confidence
                     (send ?sign1 get-confidence)
                     (send ?sign2 get-confidence)
                     (* (- 1 (/ (abs (- (send ?dgemm get-timestamp)
                                        (send ?process get-timestamp)))
                                600)) 100)))
  (bind ?severity (combine-severity (send ?sign1 get-severity)
                                    (send ?sign2 get-severity)))
  (make-instance of DIAGNOSIS (node_id ?node) (confidence ?confidence)
                 (severity ?severity)
                 (id "substandard-dgemm-due-to-high-memory-process")
                 (args (create$ (/ (send ?dgemm get-flops) 1e9)
                                (send ?process get-pid)))
                 (remedy "process-is-high-memory-remedy")
                 (remedy-args (create$ (send ?process get-pid))))
  (modify-instance ?sign1 (state diagnosed))
  (modify-instance ?sign2 (state diagnosed)))

(defrule substandard-dgemm-due-to-offline-cores
  "Diagnose substandard DGEMM performance issues due to  detected offline
   cores.  If the dgemm-performance sign is substandard and the
   all-logical-cores-not-available sign is true and the associated data
   points are close together in time (within 10 minutes), then conclude
   the substandard performance may be due to the offline cores."
  (object (is-a CONFIG) (name [config]) (data-source-time-difference ?tdiff))
  ?sign1 <- (object (is-a SIGN) (node_id ?node)
                    (id "dgemm-data-is-substandard") (source ?dgemm)
                    (suppressed FALSE) (value substandard))
  ?sign2 <- (object (is-a SIGN) (node_id ?node) (args $?args)
                    (id "all-logical-cores-not-available") (source ?cpu)
                    (suppressed FALSE) (value TRUE))
  (test (< (abs (- (send ?dgemm get-timestamp)
                   (send ?cpu get-timestamp))) ?tdiff))
  =>
  ; the further apart the two data points in time, the lower the confidence
  (bind ?confidence (combine-confidence
                     (send ?sign1 get-confidence)
                     (send ?sign2 get-confidence)))
  (bind ?severity (combine-severity (send ?sign1 get-severity)
                                    (send ?sign2 get-severity)))
  (make-instance of DIAGNOSIS (node_id ?node) (confidence ?confidence)
                 (severity ?severity)
                 (id "substandard-dgemm-due-to-offline-cores")
                 (args (create$ (/ (send ?dgemm get-flops) 1e9)
                                (send ?cpu get-offline-logical-cores)))
                 (remedy "all-logical-cores-not-available-remedy"))
  (modify-instance ?sign1 (state diagnosed))
  (modify-instance ?sign2 (state diagnosed)))

(defrule substandard-dgemm-due-to-dimms
    "Diagnose substandard DGEMM performance issues due to insufficient
     DIMMs.  If the dgemm-performance sign is substandard and the DIMMs per
     socket is insufficent"
    (object (is-a CONFIG) (name [config]) (dgemm-peak-fraction ?fraction))
    ?sign1 <- (object (is-a SIGN) (node_id ?node)
                      (id "dgemm-data-is-substandard") (source ?dgemm) (state initialized)
                      (suppressed FALSE) (value substandard))
    ?mem <- (object (is-a MEMORY) (node_id ?node) (min-dimms-per-socket ?mdps))
    (test (> ?mdps 0))
    =>
    ; set the state to observed
    (bind ?nstate observed)
    (bind ?confidence (send ?sign1 get-confidence))
    (bind ?severity (send ?sign1 get-severity))
    (bind ?pfraction (send ?dgemm get-peak-fraction))
    ; DIMMs per socket less than equal to 4
    (if (<= ?mdps 4)  then
       ; set the state to diagnosed
       (bind ?nstate diagnosed)
       (make-instance of DIAGNOSIS (node_id ?node) (confidence ?confidence)
                      (severity ?severity)
                      (id "substandard-dgemm-due-to-dimms")
                      (args (create$ (/ (send ?dgemm get-flops) 1e9)
                                     (* (send ?dgemm get-pk-fraction ?fraction ?pfraction) 100)
                                     (/ (send ?dgemm get-peak) 1e9)
                                     ?mdps))
                      (remedy "substandard-dgemm-due-to-dimms-remedy")
                      (remedy-args (create$ ?mdps))))
    (modify-instance ?sign1 (state ?nstate)))

(defrule substandard-sgemm-due-to-high-cpu-process
  "Diagnose substandard SGEMM performance issues due to a conflicting
   process that is consuming a high amount of cpu.  If the
   sgemm-performance sign is substandard and the high-cpu-process sign
   is true and the associated data points are close together in
   time (within 10 minutes), then conclude the substandard performance
   is due to the high cpu process."
  ?sign1 <- (object (is-a SIGN) (node_id ?node)
                    (id "sgemm-data-is-substandard") (source ?sgemm)
                    (suppressed FALSE) (value substandard))
  ?sign2 <- (object (is-a SIGN) (node_id ?node) (id "process-is-high-cpu")
                    (source ?process) (suppressed FALSE) (value TRUE))
  (test (< (abs (- (send ?sgemm get-timestamp)
                   (send ?process get-timestamp))) 600))
  =>
  ; the further apart the two data points in time, the lower the confidence
  (bind ?confidence (combine-confidence
                     (send ?sign1 get-confidence)
                     (send ?sign2 get-confidence)
                     (* (- 1 (/ (abs (- (send ?sgemm get-timestamp)
                                        (send ?process get-timestamp)))
                                600)) 100)))
  (bind ?severity (combine-severity (send ?sign1 get-severity)
                                    (send ?sign2 get-severity)))
  (make-instance of DIAGNOSIS (node_id ?node) (confidence ?confidence)
                 (severity ?severity)
                 (id "substandard-sgemm-due-to-high-cpu-process")
                 (args (create$ (/ (send ?sgemm get-flops) 1e9)
                                (send ?process get-pid)))
                 (remedy "process-is-high-cpu-remedy")
                 (remedy-args (create$ (send ?process get-pid))))
  (modify-instance ?sign1 (state diagnosed))
  (modify-instance ?sign2 (state diagnosed)))

(defrule substandard-sgemm-due-to-high-memory-process
  "Diagnose substandard SGEMM performance issues due to a conflicting
   process that is consuming a large amount of memory.  If the
   sgemm-performance sign is substandard and the high-memory-process
   sign is true and the associated data points are close together in
   time (within 10 minutes), then conclude the substandard performance
   is due to the high memory process."
  ?sign1 <- (object (is-a SIGN) (node_id ?node)
                    (id "sgemm-data-is-substandard") (source ?sgemm)
                    (suppressed FALSE) (value substandard))
  ?sign2 <- (object (is-a SIGN) (node_id ?node) (args $?args)
                    (id "process-is-high-memory") (source ?process)
                    (suppressed FALSE) (value TRUE))
  (test (< (abs (- (send ?sgemm get-timestamp)
                   (send ?process get-timestamp))) 600))
  =>
  ; the further apart the two data points in time, the lower the confidence
  (bind ?confidence (combine-confidence
                     (send ?sign1 get-confidence)
                     (send ?sign2 get-confidence)
                     (* (- 1 (/ (abs (- (send ?sgemm get-timestamp)
                                        (send ?process get-timestamp)))
                                600)) 100)))
  (bind ?severity (combine-severity (send ?sign1 get-severity)
                                    (send ?sign2 get-severity)))
  (make-instance of DIAGNOSIS (node_id ?node) (confidence ?confidence)
                 (severity ?severity)
                 (id "substandard-sgemm-due-to-high-memory-process")
                 (args (create$ (/ (send ?sgemm get-flops) 1e9)
                                (send ?process get-pid)))
                 (remedy "process-is-high-memory-remedy")
                 (remedy-args (create$ (send ?process get-pid))))
  (modify-instance ?sign1 (state diagnosed))
  (modify-instance ?sign2 (state diagnosed)))

(defrule substandard-sgemm-due-to-offline-cores
  "Diagnose substandard SGEMM performance issues due to  detected offline
   cores.  If the sgemm-performance sign is substandard and the
   all-logical-cores-not-available sign is true and the associated data
   points are close together in time (within 10 minutes), then conclude
   the substandard performance may be due to the offline cores."
  (object (is-a CONFIG) (name [config]) (data-source-time-difference ?tdiff))
  ?sign1 <- (object (is-a SIGN) (node_id ?node)
                    (id "sgemm-data-is-substandard") (source ?sgemm)
                    (suppressed FALSE) (value substandard))
  ?sign2 <- (object (is-a SIGN) (node_id ?node) (args $?args)
                    (id "all-logical-cores-not-available") (source ?cpu)
                    (suppressed FALSE) (value TRUE))
  (test (< (abs (- (send ?sgemm get-timestamp)
                   (send ?cpu get-timestamp))) ?tdiff))
  =>
  ; the further apart the two data points in time, the lower the confidence
  (bind ?confidence (combine-confidence
                     (send ?sign1 get-confidence)
                     (send ?sign2 get-confidence)))
  (bind ?severity (combine-severity (send ?sign1 get-severity)
                                    (send ?sign2 get-severity)))
  (make-instance of DIAGNOSIS (node_id ?node) (confidence ?confidence)
                 (severity ?severity)
                 (id "substandard-sgemm-due-to-offline-cores")
                 (args (create$ (/ (send ?sgemm get-flops) 1e9)
                                (send ?cpu get-offline-logical-cores)))
                 (remedy "all-logical-cores-not-available-remedy"))
  (modify-instance ?sign1 (state diagnosed))
  (modify-instance ?sign2 (state diagnosed)))

(defrule substandard-imb-pingpong-latency-due-to-ethernet-coalescing
  "Diagnose substandard IMB pingpong latency performance issues due to
   Ethernet interrupt coalescing not being disabled.  If the
   imb-pingpong-latency-threshold sign is TRUE (substandard), the
   Intel(R) MPI Library settings are configured to use Ethernet, and
   the ethernet-interrupt-coalescing-is-enabled sign is TRUE, then
   conclude the substandard performance is due to Ethernet interrupt
   coalescing not being disabled.  Note that the Ethernet interrupt
   coalescing only affects IMB pingpong latency, not bandwidth, so
   there is no corresponding rule for bandwidth."

  ;TODO: update to be like the outlier rule
  ?lat-sign <- (object (is-a SIGN)
                       (id "imb-pingpong-latency-threshold")
                       (source ?pair) (suppressed FALSE) (value TRUE))
  ?eth-sign <- (object (is-a SIGN)
                       (id "ethernet-interrupt-coalescing-is-enabled")
                       (source ?node) (suppressed FALSE) (value TRUE))
  (object (is-a IMB_PINGPONG) (i_mpi_fabrics ?f&:(str-index "tcp" ?f))
          (i_mpi_tcp_netmask ?m&:(str-index "eth" ?m))
          (name ?n&:(eq ?n (instance-name ?pair))))
  =>
  (bind ?confidence (combine-confidence (send ?eth-sign get-confidence)
                                        (send ?lat-sign get-confidence)))
  (bind ?severity (combine-severity (send ?eth-sign get-severity)
                                    (send ?lat-sign get-severity)))
  (make-instance of DIAGNOSIS (node_id (send ?pair get-node_id))
		(confidence ?confidence) (severity ?severity)
        (id "substandard-imb-pingpong-latency-due-to-ethernet-coalescing")
        (args (create$ (send ?pair get-latency)
                            (send ?pair get-i_mpi_tcp_netmask)
                            (send ?pair get-impi-setting-string)))
        (remedy "substandard-imb-pingpong-latency-due-to-ethernet-coalescing-remedy")
        (remedy-args (create$ (send ?node get-interface))))
  (modify-instance ?eth-sign (state diagnosed))
  (modify-instance ?lat-sign (state diagnosed)))

(defrule outlier-imb-pingpong-latency-due-to-ethernet-coalescing
  "Diagnose IMB pingpong latency performance outlier issues due to
   Ethernet interrupt coalescing not being disabled.  If the
   imb-pingpong-latency-outlier sign is TRUE, the
   Intel(R) MPI Library settings are configured to use Ethernet, and
   the ethernet-interrupt-coalescing-is-enabled sign is TRUE, then
   conclude the inconsistent performance is due to Ethernet interrupt
   coalescing not being disabled.  Note that the Ethernet interrupt
   coalescing only affects IMB pingpong latency, not bandwidth, so
   there is no corresponding rule for bandwidth."
  ?eth-sign <- (object (is-a SIGN) (node_id ?n)
                       (id "ethernet-interrupt-coalescing-is-enabled")
                       (source ?node) (suppressed FALSE) (value TRUE))
  ?lat-sign <- (object (is-a SIGN) (node_id $? ?n $?)
                       (id "imb-pingpong-latency-outlier")
                       (source ?pair) (suppressed FALSE) (value TRUE))
  (object (is-a IMB_PINGPONG) (i_mpi_fabrics ?f&:(str-index "tcp" ?f))
          (i_mpi_tcp_netmask ?m&:(str-index "eth" ?m))
          (name ?name&:(eq ?name (instance-name ?pair))))
  (not (object (is-a DIAGNOSIS) (node_id ?n)
               (id "outlier-imb-pingpong-latency-due-to-ethernet-coalescing")))
  =>
  (bind ?confidence (combine-confidence (send ?eth-sign get-confidence)
                                        (send ?lat-sign get-confidence)))
  (bind ?severity (combine-severity (send ?eth-sign get-severity)
                                    (send ?lat-sign get-severity)))
  (make-instance of DIAGNOSIS (node_id ?n)
                 (confidence ?confidence) (severity ?severity)
                 (id "outlier-imb-pingpong-latency-due-to-ethernet-coalescing")
                 (args (create$ (send ?node get-interface)))
                 (remedy "outlier-imb-pingpong-latency-due-to-ethernet-coalescing-remedy")
                 (remedy-args (create$ (send ?node get-interface))))
  (modify-instance ?eth-sign (state diagnosed))
  (modify-instance ?lat-sign (state diagnosed)))

