#!/bin/bash
# Copyright 2015-2019 Intel Corporation
#
# This software and the related documents are Intel copyrighted materials, and
# your use of them is governed by the express license under which they were
# provided to you (License). Unless the License provides otherwise, you may
# not use, modify, copy, publish, distribute, disclose, or transmit this
# software or the related documents without Intel's prior written permission.
#
# This software and the related documents are provided as is, with no express
# or implied warranties, other than those that are expressly stated in the
# License.

. $(dirname $0)/../../functions

ARCH=`uname -m`

# set the number of iterations for dgemm wrapper
function set_iterations() {
    clck_dgemm_iterations=$1
    if [[ ! -z $CLCK_PROVIDER_DGEMM_ITERATIONS ]] ; then
        if [[ $CLCK_PROVIDER_DGEMM_ITERATIONS =~ ^[0-9]+$ && $CLCK_PROVIDER_DGEMM_ITERATIONS -gt 0 ]] ; then
            clck_dgemm_iterations=$CLCK_PROVIDER_DGEMM_ITERATIONS
        fi
    fi
}

# sets the memory usage parameters
function set_memory_usage() {
    clck_mu_rtype=1
    clck_dgemm_memory_usage=$1
    k_param=$2
    if [[ ! -z $CLCK_PROVIDER_DGEMM_MEMORY_USAGE && \
               $CLCK_PROVIDER_DGEMM_MEMORY_USAGE =~ ^[0-9]+$ ]] ; then
        if [[ $CLCK_PROVIDER_DGEMM_MEMORY_USAGE -gt 0 && \
            $CLCK_PROVIDER_DGEMM_MEMORY_USAGE -le 100 ]] ; then
            clck_dgemm_memory_usage=${CLCK_PROVIDER_DGEMM_MEMORY_USAGE}
            clck_mu_rtype=0
            if [[ ! -z $CLCK_PROVIDER_DGEMM_K_PARAMETER ]] ; then
                 if [[ $CLCK_PROVIDER_DGEMM_K_PARAMETER =~ ^[0-9]+$ && \
                       $CLCK_PROVIDER_DGEMM_K_PARAMETER -gt 0 ]] ; then
                     k_param=${CLCK_PROVIDER_DGEMM_K_PARAMETER}
                     clck_mu_rtype=0
                 else
                     clck_mu_rtype=2
                     _stderr "invalid k parameter , using the default value"
                 fi
            fi
            return $clck_mu_rtype
        else
            _stderr "invalid memory usage percentage, using the default value"
        fi
    fi
    return $clck_mu_rtype
}

function set_m_n_k() {
    m_param=$1
    n_param=$2
    k_param=$3

    # memory percentage was set, do not over-ride
    if [[ $clck_mu_rtype == 0 ]] ; then
        return;
    fi

    if [[ ! -z $CLCK_PROVIDER_DGEMM_M_PARAMETER && ! -z $CLCK_PROVIDER_DGEMM_N_PARAMETER && ! -z $CLCK_PROVIDER_DGEMM_K_PARAMETER  ]] ; then
          if [[ ( $CLCK_PROVIDER_DGEMM_M_PARAMETER =~ ^[0-9]+$ && $CLCK_PROVIDER_DGEMM_M_PARAMETER -gt 0 ) && \
              ( $CLCK_PROVIDER_DGEMM_N_PARAMETER =~ ^[0-9]+$ && $CLCK_PROVIDER_DGEMM_N_PARAMETER -gt 0 ) && \
               ( $CLCK_PROVIDER_DGEMM_K_PARAMETER =~ ^[0-9]+$ && $CLCK_PROVIDER_DGEMM_K_PARAMETER -gt 0 ) ]] ; then
            m_param=${CLCK_PROVIDER_DGEMM_M_PARAMETER}
            n_param=${CLCK_PROVIDER_DGEMM_N_PARAMETER}
            k_param=${CLCK_PROVIDER_DGEMM_K_PARAMETER}
        fi
    fi
}

function set_kmp_affinity() {
    if [[ "$CLCK_PROVIDER_DGEMM_KMP_AFFINITY" == "" ]]; then
        export KMP_AFFINITY="$1"
    else
        export KMP_AFFINITY="${CLCK_PROVIDER_DGEMM_KMP_AFFINITY}"
    fi
}

function set_kmp_hw_subset() {
    if [[ "$CLCK_PROVIDER_DGEMM_KMP_HW_SUBSET" == "" ]]; then
        export KMP_HW_SUBSET="$1"
    else
        export KMP_HW_SUBSET="${CLCK_PROVIDER_DGEMM_KMP_HW_SUBSET}"
    fi
}

function get_task_set() {
    # number of supported cores and threads/core
    local taskset_threads=$(( $1*$2 ))
    taskset_threads=$(( taskset_threads-1 ))
    taskset_list="${taskset_threads}"
    taskset_list="0-${taskset_list}"
    if [[ "$CLCK_PROVIDER_DGEMM_TASKSET" != "" ]]; then
        taskset_list="${CLCK_PROVIDER_DGEMM_TASKSET}"
    fi
}

function clck_detect_taskset() {
    get_tool_presence "taskset"
    clck_taskset_bin=$tool_path
    if [[ "$CLCK_PROVIDER_DGEMM_TASKSET_BINARY" != "" ]]; then
        clck_taskset_bin="${CLCK_PROVIDER_DGEMM_TASKSET_BINARY}"
    fi
    # number of supported cores and threads/core
    num_physical_cores
    num_threads_per_core
    get_task_set ${num_physical_cores} ${num_threads_per_core}
    if [[ $clck_taskset_bin == "CLCK_LOCATION_NOT_FOUND" ]]; then
        clck_taskset_bin="";
    else
        clck_taskset_bin="$clck_taskset_bin -c ${taskset_list}"
    fi
}

# prepare taskset command
clck_detect_taskset

case "${ARCH}" in
    "x86_64")
        export LD_LIBRARY_PATH=$(dirname $0)/../common/lib/intel64
        get_cpu_type
        num_physical_cores
        num_threads_per_core
        # check the cpu type for Intel(R) Xeon Phi(TM) processor x200 product family,
        if [[ $clck_cpu_type =~ ^xeonphi* ]]; then
           m=28000
           n=28000
           k=288
           iterations=9
           clck_mu_rtype=1

           set_iterations $iterations
           set_m_n_k $m $n $k
           set_kmp_affinity "scatter"
           set_kmp_hw_subset "${num_physical_cores}c,${num_threads_per_core}t"
           exec $clck_taskset_bin $(dirname $0)/intel64/dgemm $m_param $n_param $k_param $clck_dgemm_iterations 1
        else
           mem_percent=20
           iterations=9

           if [[ $clck_cpu_type == "xeon_scalable" ]]; then
             k=288
           else
             k=240
           fi

           set_memory_usage $mem_percent $k

           memory_size 'KB'

           msq=$(( ($memory_size * 1024 * $clck_dgemm_memory_usage / 800) + ($k_param * $k_param) - $k_param ))
           mi=`echo $msq | awk '{ printf("%d", sqrt($1)) }'`

           if (( $mi % 128 > 64 )); then
               m=$(( $mi + 128 - $mi % 128 ))
           else
               m=$(( $mi - $mi % 128 ))
           fi

           n=$m

           set_iterations $iterations
           set_m_n_k $m $n $k_param
           set_kmp_affinity "granularity=fine,compact,1,0"
           exec $clck_taskset_bin $(dirname $0)/intel64/dgemm $m_param $n_param $k_param $clck_dgemm_iterations 1
        fi
        ;;

    *)
        echo "Error executing dgemm, node arch unknown. (${ARCH})"
        exit 1
        ;;
esac
